@@ -154,6 +154,16 @@ static void handle_run_command        (MetaDisplay    *display,
                                        MetaWindow     *window,
                                        XEvent         *event,
                                        MetaKeyBinding *binding);
+static void handle_maximize_vert      (MetaDisplay    *display,
+                                       MetaScreen     *screen,
+                                       MetaWindow     *window,
+                                       XEvent         *event,
+                                       MetaKeyBinding *binding);
+static void handle_maximize_horiz     (MetaDisplay    *display,
+                                       MetaScreen     *screen,
+                                       MetaWindow     *window,
+                                       XEvent         *event,
+                                       MetaKeyBinding *binding);
 
 /* debug */
 static void handle_spew_mark          (MetaDisplay *display,
@@ -329,6 +339,8 @@ static const MetaKeyHandler window_handlers[] = {
   { META_KEYBINDING_RAISE_OR_LOWER, handle_raise_or_lower, NULL},
   { META_KEYBINDING_RAISE, handle_raise, NULL},
   { META_KEYBINDING_LOWER, handle_lower, NULL},
+  { META_KEYBINDING_MAXIMIZE_VERTICALLY, handle_maximize_vert, NULL },
+  { META_KEYBINDING_MAXIMIZE_HORIZONTALLY, handle_maximize_horiz, NULL },
   { NULL, NULL, NULL }
 };
 
@@ -395,7 +407,7 @@ reload_modmap (MetaDisplay *display)
                   str = XKeysymToString (syms[j]);
                   meta_topic (META_DEBUG_KEYBINDINGS,
                               "Keysym %s bound to modifier 0x%x\n",
-                              str ? str : "(null)",
+                              str ? str : "none",
                               (1 << ( i / modmap->max_keypermod)));
                 }
               
@@ -1368,7 +1380,7 @@ meta_display_process_key_event (MetaDisplay *display,
   meta_topic (META_DEBUG_KEYBINDINGS,
               "Processing key %s event, keysym: %s state: 0x%x window: %s\n",
               event->type == KeyPress ? "press" : "release",
-              str ? str : "(null)", event->xkey.state,
+              str ? str : "none", event->xkey.state,
               window ? window->desc : "(no window)");
 
   all_keys_grabbed = window ? window->all_keys_grabbed : screen->all_keys_grabbed;
@@ -2239,6 +2251,35 @@ handle_run_command (MetaDisplay    *display,
     }
 }
 
+
+static void
+handle_maximize_vert (MetaDisplay    *display,
+                      MetaScreen     *screen,
+                      MetaWindow     *window,
+                      XEvent         *event,
+                      MetaKeyBinding *binding)
+{
+  if (window)
+    {
+      if (window->has_resize_func)
+        meta_window_fill_vertical (window);
+    }
+}
+
+static void
+handle_maximize_horiz (MetaDisplay    *display,
+                       MetaScreen     *screen,
+                       MetaWindow     *window,
+                       XEvent         *event,
+                       MetaKeyBinding *binding)
+{
+  if (window)
+    {
+      if (window->has_resize_func)
+        meta_window_fill_horizontal (window);
+    }
+}
+
 static gboolean
 process_workspace_switch_grab (MetaDisplay *display,
                                MetaScreen  *screen,
