@@ -87,6 +87,7 @@ from The Open Group.
 #include "metacity-Xatomtype.h"
 #include <X11/Xatom.h>
 #include <string.h>
+#include "window.h"
 
 typedef struct
 {
@@ -109,7 +110,11 @@ validate_or_free_results (GetPropertyResults *results,
   char *type_name;
   char *expected_name;
   char *prop_name;
-
+  const char *title;
+  const char *res_class;
+  const char *res_name;
+  MetaWindow *w;
+  
   if (expected_format == results->format &&
       expected_type == results->type &&
       (!must_have_items || results->n_items > 0))
@@ -121,13 +126,38 @@ validate_or_free_results (GetPropertyResults *results,
   prop_name = XGetAtomName (results->display->xdisplay, results->xatom);
   meta_error_trap_pop (results->display, TRUE);
 
-  meta_warning (_("Window 0x%lx has property %s that was expected to have type %s format %d and actually has type %s format %d n_items %d\n"),
+  w = meta_display_lookup_x_window (results->display, results->xwindow);
+
+  if (w != NULL)
+    {
+      title = w->title;
+      res_class = w->res_class;
+      res_name = w->res_name;
+    }
+  else
+    {
+      title = NULL;
+      res_class = NULL;
+      res_name = NULL;
+    }
+  
+  if (title == NULL)
+    title = "unknown";
+
+  if (res_class == NULL)
+    res_class = "unknown";
+
+  if (res_name == NULL)
+    res_name = "unknown";
+  
+  meta_warning (_("Window 0x%lx has property %s\nthat was expected to have type %s format %d\nand actually has type %s format %d n_items %d.\nThis is most likely an application bug, not a window manager bug.\nThe window has title=\"%s\" class=\"%s\" name=\"%s\"\n"),
                 results->xwindow,
                 prop_name ? prop_name : "(bad atom)",
                 expected_name ? expected_name : "(bad atom)",
                 expected_format,
                 type_name ? type_name : "(bad atom)",
-                results->format, (int) results->n_items);
+                results->format, (int) results->n_items,
+                title, res_class, res_name);
 
   if (type_name)
     XFree (type_name);
@@ -505,6 +535,24 @@ window_from_results (GetPropertyResults *results,
   return TRUE;
 }
 
+#ifdef HAVE_XSYNC
+static gboolean
+counter_from_results (GetPropertyResults *results,
+                      XSyncCounter       *counter_p)
+{
+  if (!validate_or_free_results (results, 32,
+                                 results->display->atom_sync_counter,
+                                 TRUE))
+    return FALSE;  
+
+  *counter_p = *(XSyncCounter*) results->prop;
+  XFree (results->prop);
+  results->prop = NULL;
+  
+  return TRUE;
+}
+#endif
+
 gboolean
 meta_prop_get_window (MetaDisplay *display,
                       Window       xwindow,
@@ -849,6 +897,24 @@ get_task (MetaDisplay        *display,
                          False, req_type);
 }
 
+static char*
+latin1_to_utf8 (const char *text)
+{
+  GString *str;
+  const char *p;
+  
+  str = g_string_new ("");
+
+  p = text;
+  while (*p)
+    {
+      g_string_append_unichar (str, *p);
+      ++p;
+    }
+
+  return g_string_free (str, FALSE);
+}
+
 void
 meta_prop_get_values (MetaDisplay   *display,
                       Window         xwindow,
@@ -888,6 +954,7 @@ meta_prop_get_values (MetaDisplay   *display,
               values[i].required_type = display->atom_utf8_string;
               break;
             case META_PROP_VALUE_STRING:
+            case META_PROP_VALUE_STRING_AS_UTF8:
               values[i].required_type = XA_STRING;
               break;
             case META_PROP_VALUE_MOTIF_HINTS:
@@ -915,6 +982,9 @@ meta_prop_get_values (MetaDisplay   *display,
             case META_PROP_VALUE_SIZE_HINTS:
               values[i].required_type = XA_WM_SIZE_HINTS;
               break;
+            case META_PROP_VALUE_SYNC_COUNTER:
+              values[i].required_type = display->atom_sync_counter;
+              break;
             }
         }
 
@@ -996,6 +1066,27 @@ meta_prop_get_values (MetaDisplay   *display,
                                            &values[i].v.str))
             values[i].type = META_PROP_VALUE_INVALID;
           break;
+        case META_PROP_VALUE_STRING_AS_UTF8:
+          if (!latin1_string_from_results (&results,
+                                           &values[i].v.str))
+            values[i].type = META_PROP_VALUE_INVALID;
+          else
+            {
+              char *new_str;
+              char *xmalloc_new_str;
+
+              new_str = latin1_to_utf8 (values[i].v.str);
+              xmalloc_new_str = ag_Xmalloc (strlen (new_str) + 1);
+              if (xmalloc_new_str != NULL)
+                {
+                  strcpy (xmalloc_new_str, new_str);
+                  meta_XFree (values[i].v.str);
+                  values[i].v.str = xmalloc_new_str;
+                }
+
+              g_free (new_str);
+            }
+          break;
         case META_PROP_VALUE_MOTIF_HINTS:
           if (!motif_hints_from_results (&results,
                                          &values[i].v.motif_hints))
@@ -1042,6 +1133,15 @@ meta_prop_get_values (MetaDisplay   *display,
                                         &values[i].v.size_hints.flags))
             values[i].type = META_PROP_VALUE_INVALID;
           break;
+        case META_PROP_VALUE_SYNC_COUNTER:
+#ifdef HAVE_XSYNC
+          if (!counter_from_results (&results,
+                                     &values[i].v.xcounter))
+            values[i].type = META_PROP_VALUE_INVALID;
+#else
+          values[i].type = META_PROP_VALUE_INVALID;
+#endif
+          break;
         }
       
     next:
@@ -1059,9 +1159,8 @@ free_value (MetaPropValue *value)
     case META_PROP_VALUE_INVALID:          
       break;
     case META_PROP_VALUE_UTF8:
-      meta_XFree (value->v.str);
-      break;
     case META_PROP_VALUE_STRING:
+    case META_PROP_VALUE_STRING_AS_UTF8:
       meta_XFree (value->v.str);
       break;
     case META_PROP_VALUE_MOTIF_HINTS:
@@ -1093,6 +1192,8 @@ free_value (MetaPropValue *value)
     case META_PROP_VALUE_CARDINAL_LIST:
       meta_XFree (value->v.cardinal_list.cardinals);
       break;
+    case META_PROP_VALUE_SYNC_COUNTER:
+      break;
     }
 }
 
